from flask import (abort, current_app, flash, redirect, render_template,
                   send_from_directory, url_for)
from flask_login import current_user, login_required
from . import jobs
from . import tasks
from ..decorators import admin_required
from ..models import Job, JobInput, JobResult
import os


@jobs.route('/<int:job_id>')
@login_required
def job(job_id):
    job = Job.query.get_or_404(job_id)
    if not (job.creator == current_user or current_user.is_administrator()):
        abort(403)
    job_inputs = [dict(filename=input.filename,
                       id=input.id,
                       job_id=job.id)
                  for input in job.inputs]
    return render_template('jobs/job.html.j2',
                           job=job,
                           job_inputs=job_inputs,
                           title='Job')


@jobs.route('/<int:job_id>/delete')
@login_required
def delete_job(job_id):
    job = Job.query.get_or_404(job_id)
    if not (job.creator == current_user or current_user.is_administrator()):
        abort(403)
    tasks.delete_job(job_id)
    flash('Job has been deleted!', 'job')
    return redirect(url_for('main.dashboard'))


@jobs.route('/<int:job_id>/inputs/<int:job_input_id>/download')
@login_required
def download_job_input(job_id, job_input_id):
    job_input = JobInput.query.get_or_404(job_input_id)
    if not job_input.job_id == job_id:
        abort(404)
    if not (job_input.job.creator == current_user
            or current_user.is_administrator()):
        abort(403)
    dir = os.path.join(current_app.config['NOPAQUE_STORAGE'],
                       job_input.dir)
    return send_from_directory(as_attachment=True, directory=dir,
                               filename=job_input.filename)


@jobs.route('/<int:job_id>/restart')
@login_required
@admin_required
def restart(job_id):
    job = Job.query.get_or_404(job_id)
    if job.status != 'failed':
        flash('Could not restart job: status is not "failed"', 'error')
    else:
        tasks.restart_job(job_id)
        flash('Job has been restarted!', 'job')
    return redirect(url_for('jobs.job', job_id=job_id))


@jobs.route('/<int:job_id>/results/<int:job_result_id>/download')
@login_required
def download_job_result(job_id, job_result_id):
    job_result = JobResult.query.get_or_404(job_result_id)
    if not job_result.job_id == job_id:
        abort(404)
    if not (job_result.job.creator == current_user
            or current_user.is_administrator()):
        abort(403)
    dir = os.path.join(current_app.config['NOPAQUE_STORAGE'],
                       job_result.dir)
    return send_from_directory(as_attachment=True, directory=dir,
                               filename=job_result.filename)
